import asyncio
from abc import ABC
from typing import Union, Callable, Awaitable, Iterable

from lcu_driver.events.responses import WebsocketEventResponse


class ConnectorEventManager(ABC):
    """Connector Events Manager Base Class"""

    def __init__(self):
        self._handlers = {}

    @property
    def handlers(self):
        return self._handlers

    def _set_event(self, event_name, func_or_coro: Union[Callable, Awaitable]):
        if event_name in self._handlers:
            self.handlers[event_name].append(func_or_coro)
        else:
            self.handlers[event_name] = [func_or_coro, ]
        return self.handlers[event_name][-1]

    async def run_event(self, event_name, *args, **kwargs):
        for event in self._handlers.get(event_name, []):
            await asyncio.create_task(
                event(*args, **kwargs)
            )

    def open(self, coro_func):
        return self._set_event('open', coro_func)

    def ready(self, coro_func):
        return self._set_event('ready', coro_func)

    def close(self, coro_func):
        return self._set_event('close', coro_func)


class WebsocketEventManager(ABC):
    """Connector Events Manager Base Class"""

    def __init__(self):
        self._registered_uris = []

    @property
    def registered_uris(self) -> list:
        """Websocket registered handlers

        :rtype: list
        """
        return self._registered_uris

    def register(self, uri: str, *, event_types: Iterable = ('CREATE', 'UPDATE', 'DELETE',)):
        """Register an event for the given handler.

        :param string uri: Endpoint to call. If the endpoint last character is a slash it will match all events starting with the endpoint.
        :param event_types: Expects an iterable. The allowed types are CREATE, UPDATE and DELETE (case-sensitive).
        :type event_types: tuple(str, str)
        """
        allowed_events = ('CREATE', 'UPDATE', 'DELETE',)

        if not uri.startswith('/'):
            raise RuntimeError('every endpoint should start with backslash')

        def register_wrapper(coro_func):
            for event in event_types:
                if event not in allowed_events:
                    raise RuntimeError(f'Event {event} not recognized.')

            self._registered_uris.append({
                'uri': uri,
                'event_types': event_types,
                'coroutine_or_callable': coro_func
            })
            return coro_func
        return register_wrapper

    @staticmethod
    def match_event(connector, connection, data):
        """Match registered websocket events and create a task with each handler"""
        for event in connector.ws.registered_uris:
            if event['uri'] == data['uri'] or (
                    event['uri'].endswith('/') and data['uri'].startswith(event['uri'])
            ):
                if data['eventType'].upper() in event['event_types']:
                    ws_dto = WebsocketEventResponse(
                        event_type=data['eventType'],
                        uri=data['uri'],
                        data=data['data'],
                    )
                    asyncio.create_task(event['coroutine_or_callable'](connection, ws_dto))



import asyncio
import logging
import time
from abc import ABC, abstractmethod

from .connection import Connection
from .events.managers import ConnectorEventManager, WebsocketEventManager
from .utils import _return_ux_process

logger = logging.getLogger('lcu-driver')


class BaseConnector(ConnectorEventManager, ABC):
    def __init__(self, loop=None):
        super().__init__()
        self.loop = loop or asyncio.get_event_loop()
        self.ws = WebsocketEventManager()

    @abstractmethod
    def register_connection(self, connection: Connection):
        """Creates a connection and saves a reference to it"""
        pass

    @abstractmethod
    def unregister_connection(self, lcu_pid):
        """Cancel the connection"""
        pass

    @property
    def should_run_ws(self) -> bool:
        return True


class Connector(BaseConnector):
    def __init__(self, *, loop=None):
        super().__init__(loop)
        self._repeat_flag = True
        self.connection = None

    def register_connection(self, connection):
        self.connection = connection

    def unregister_connection(self, _):
        self.connection = None

    @property
    def should_run_ws(self) -> bool:
        return len(self.ws.registered_uris) > 0

    def start(self) -> None:
        """Starts the connector. This method should be overridden if different behavior is required.

        :rtype: None
        """
        try:
            def wrapper():
                process = next(_return_ux_process(), None)
                while not process:
                    process = next(_return_ux_process(), None)
                    time.sleep(0.5)

                connection = Connection(self, process)
                self.register_connection(connection)
                self.loop.run_until_complete(connection.init())

                if self._repeat_flag and len(self.ws.registered_uris) > 0:
                    logger.debug('Repeat flag=True. Looking for new clients.')
                    wrapper()

            wrapper()
        except KeyboardInterrupt:
            logger.info('Event loop interrupted by keyboard')

    async def stop(self) -> None:
        """Flag the connector to don't look for more clients once the connection finishes his job.

        :rtype: None
        """
        self._repeat_flag = False


class MultipleClientConnector(BaseConnector):
    def __init__(self, *, loop=None):
        super().__init__(loop=loop)
        self.connections = []

    def register_connection(self, connection):
        self.connections.append(connection)

    def unregister_connection(self, lcu_pid):
        for index, connection in enumerate(self.connections):
            if connection.pid == lcu_pid:
                del connection[index]

    @property
    def should_run_ws(self) -> bool:
        return True

    def _process_was_initialized(self, non_initialized_connection):
        for connection in self.connections:
            if non_initialized_connection.pid == connection.pid:
                return True
        return False

    async def _astart(self):
        tasks = []
        try:
            while True:
                process_iter = _return_ux_process()

                process = next(process_iter, None)
                while process:
                    connection = Connection(self, process)
                    if not self._process_was_initialized(connection):
                        tasks.append(asyncio.create_task(connection.init()))

                    process = next(process_iter, None)
                await asyncio.sleep(0.5)

        except KeyboardInterrupt:
            logger.info('Event loop interrupted by keyboard')
        finally:
            await asyncio.gather(*tasks)

    def start(self) -> None:
        self.loop.run_until_complete(self._astart())
